/* strongSwan IPsec config file parser
 * Copyright (C) 2001-2002 Mathieu Lafon - Arkoon Network Security
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: confread.c,v 1.22 2006/01/07 16:54:55 as Exp $
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <freeswan.h>

#include "../pluto/constants.h"
#include "../pluto/defs.h"
#include "../pluto/log.h"

#include "keywords.h"
#include "parser.h"
#include "confread.h"
#include "interfaces.h"

typedef enum {
    ARG_NONE,
    ARG_BOOL,
    ARG_ENUM,
    ARG_UINT,
    ARG_DUR,
    ARG_STR,
    ARG_STRC,
    ARG_LST,
    ARG_LSTC
} arg_t;

typedef struct {
    arg_t   type;
    size_t  offset;
} token_info_t;

static const token_info_t token_info[] =
{
    { ARG_LST,  offsetof(starter_config_t, setup.interfaces)       },
    { ARG_LSTC, offsetof(starter_config_t, setup.klipsdebug)       },
    { ARG_LSTC, offsetof(starter_config_t, setup.plutodebug)       },
    { ARG_NONE, 0                                                  },
    { ARG_STR,  offsetof(starter_config_t, setup.prepluto)         },
    { ARG_STR,  offsetof(starter_config_t, setup.postpluto)        },
    { ARG_BOOL, offsetof(starter_config_t, setup.fragicmp)         },
    { ARG_STRC, offsetof(starter_config_t, setup.packetdefault)    },
    { ARG_BOOL, offsetof(starter_config_t, setup.hidetos)          },
    { ARG_BOOL, offsetof(starter_config_t, setup.uniqueids)        },
    { ARG_UINT, offsetof(starter_config_t, setup.overridemtu)      },
    { ARG_DUR,  offsetof(starter_config_t, setup.crlcheckinterval) },
    { ARG_BOOL, offsetof(starter_config_t, setup.cachecrls)        },
    { ARG_BOOL, offsetof(starter_config_t, setup.strictcrlpolicy)  },
    { ARG_BOOL, offsetof(starter_config_t, setup.nocrsend)         },
    { ARG_BOOL, offsetof(starter_config_t, setup.nat_traversal)    },
    { ARG_UINT, offsetof(starter_config_t, setup.keep_alive)       },
    { ARG_STR,  offsetof(starter_config_t, setup.virtual_private)  },
    { ARG_STR,  offsetof(starter_config_t, setup.pkcs11module)     },
    { ARG_BOOL, offsetof(starter_config_t, setup.pkcs11keepstate)  },
    { ARG_BOOL, offsetof(starter_config_t, setup.pkcs11proxy)      },
    { ARG_ENUM, offsetof(starter_conn_t, startup)                  }
};

static char _tmp_err[512];

static const char ike_defaults[] = "3des-sha, 3des-md5";
static const char esp_defaults[] = "3des-sha1, 3des-md5";

static void
_free_list(char **list)
{
    char **s;

    for (s = list; *s; s++)
	pfree(*s);
    pfree(list);
}

static char **
_new_list(char *value)
{
    char *val, *b, *e, *end, **ret;
    int count;

    val = value ? clone_str(value, "list value") : NULL;
    if (!val)
	return NULL;
    end = val + strlen(val);
    for (b = val, count = 0; b < end;)
    {
	for (e = b; ((*e != ' ') && (*e != '\0')); e++);
	*e = '\0';
	if (e != b)
	    count++;
	b = e + 1;
    }
    if (count == 0)
    {
	pfree(val);
	return NULL;
    }
    ret = (char **)alloc_bytes((count+1) * sizeof(char *), "list");
    if (!ret)
    {
	pfree(val);
	return NULL;
    }
    for (b = val, count = 0; b < end; )
    {
	for (e = b; (*e != '\0'); e++);
	if (e != b)
	    ret[count++] = clone_str(b, "list value");
	b = e + 1;
    }
    ret[count] = NULL;
    pfree(val);
    return ret;
}

static void
default_values(starter_config_t *cfg)
{
    if (!cfg)
	return;
    memset(cfg, 0, sizeof(struct starter_config));

    cfg->setup.fragicmp   = TRUE;
    cfg->setup.hidetos    = TRUE;
    cfg->setup.uniqueids  = TRUE;
    cfg->setup.interfaces = _new_list("%defaultroute");

    cfg->conn_default.policy = POLICY_ENCRYPT | POLICY_TUNNEL | POLICY_RSASIG
			     | POLICY_PFS;

    cfg->conn_default.ike                   = clone_str(ike_defaults, "ike_defaults");
    cfg->conn_default.esp                   = clone_str(esp_defaults, "esp_defaults");
    cfg->conn_default.sa_ike_life_seconds   = OAKLEY_ISAKMP_SA_LIFETIME_DEFAULT;
    cfg->conn_default.sa_ipsec_life_seconds = SA_LIFE_DURATION_DEFAULT;
    cfg->conn_default.sa_rekey_margin       = SA_REPLACEMENT_MARGIN_DEFAULT;
    cfg->conn_default.sa_rekey_fuzz         = SA_REPLACEMENT_FUZZ_DEFAULT;
    cfg->conn_default.sa_keying_tries       = SA_REPLACEMENT_RETRIES_DEFAULT;

    cfg->conn_default.addr_family           = AF_INET;
    cfg->conn_default.tunnel_addr_family    = AF_INET;

    anyaddr(AF_INET, &cfg->conn_default.left.addr);
    anyaddr(AF_INET, &cfg->conn_default.left.nexthop);
    anyaddr(AF_INET, &cfg->conn_default.left.srcip);
    anyaddr(AF_INET, &cfg->conn_default.right.addr);
    anyaddr(AF_INET, &cfg->conn_default.right.nexthop);
    anyaddr(AF_INET, &cfg->conn_default.right.srcip);

    cfg->conn_default.startup = STARTUP_NO;
    cfg->conn_default.state   = STATE_IGNORE;
}

#define FREE_STR(v) { if (v) { pfree(v); v = NULL; } }
#define FREE_LST(v) { if (v) { _free_list(v); v = NULL; } }

#define ERR_FOUND(args...) \
	{ if (perr && (*perr == NULL)) { \
		snprintf(_tmp_err, sizeof(_tmp_err)-1, ## args); \
		*perr = strdup(_tmp_err); } \
	err++; }

#define KW_BOOL(val) \
		if (streq(kw->value, "yes")) { val = TRUE; } \
		else if (streq(kw->value, "no")) { val = FALSE; } \
		else { ERR_FOUND("bad boolean value: %s=%s", kw->entry->name, kw->value) }

#define KW_POLICY_FLAG(sy, sn, fl) \
		if (streq(kw->value, sy)) { conn->policy |= fl; } \
		else if (streq(kw->value, sn)) { conn->policy &= ~fl; } \
		else { ERR_FOUND("bad policy value: %s=%s", kw->entry->name, kw->value) }

#define KW_POLICY_FLAG_BI(sy, sn, fy, fn) \
		if (streq(kw->value, sy)) \
			{ conn->policy |= fy; conn->policy &= ~fn; } \
		else if (streq(kw->value, sn)) \
			{ conn->policy &= ~fy; conn->policy |= fn; } \
		else { ERR_FOUND("bad policy value: %s=%s", kw->entry->name, kw->value) }

#define KW_STR(val) \
		FREE_STR(val) \
		val = clone_str(kw->value, "str_value");

#define KW_STR_IN(val, in) \
		if (!kw_in_list(kw->value, in)) \
			ERR_FOUND("bad value: %s=%s", kw->entry->name, kw->value) \
		FREE_STR(val) \
		val = clone_str(kw->value, "str_in_value");

#define KW_LIST(val) \
		FREE_LST(val) \
		val = _new_list(kw->value);

#define KW_LIST_IN(val, in) { \
		char **_priv_b; \
		FREE_LST(val) \
		val = _new_list(kw->value); \
		for (_priv_b=val; (_priv_b && *_priv_b); _priv_b++) \
			if (!kw_in_list(*_priv_b,in)) \
				ERR_FOUND("bad value: %s=%s", kw->entry->name, *_priv_b) \
	}

#define KW_INT(val) { \
		char *endptr; \
		val = strtoul(kw->value, &endptr, 10); \
		if ((*endptr != '\0') || (endptr == kw->value)) { \
			ERR_FOUND("bad integer value: %s=%s", kw->entry->name, kw->value); \
		} \
	}

#define KW_PERCENT(val) { \
		char *endptr; \
		val = strtoul(kw->value, &endptr, 10); \
		if (((*endptr != '%') || (endptr[1] != '\0')) || (endptr == kw->value)) { \
			ERR_FOUND("bad percent value: %s=%s", kw->entry->name, kw->value); \
		} \
	}

#define KW_DURATION(val) { \
		char *endptr; \
		val = strtoul(kw->value, &endptr, 10); \
		if (endptr == kw->value) { \
			ERR_FOUND("bad duration value: %s=%s", kw->entry->name, kw->value); \
		} \
		else if (*endptr == '\0') { } \
		else if ((*endptr == 's') && (endptr[1] == '\0')) { } \
		else if ((*endptr == 'm') && (endptr[1] == '\0')) { val *= 60; } \
		else if ((*endptr == 'h') && (endptr[1] == '\0')) { val *= 3600; } \
		else if ((*endptr == 'd') && (endptr[1] == '\0')) { val *= 3600*24; } \
		else { \
			ERR_FOUND("bad duration value: %s=%s", kw->entry->name, kw->value); \
		} \
	}

static int 
kw_in_list(char *val, char *in_list)
{
    char *in = clone_str(in_list, "in_list"), *b, *e, *end;

    if (!in)
	return 0;
    end = in + strlen(in);
    for (b = in; b < end; )
    {
	for (e = b; ((*e != ' ') && (*e != '\0')); e++);
	*e = '\0';
	if (e != b)
	{
	    if (streq(val, b))
	    {
		pfree(in);
		return 1;
	    }
	}
	b = e + 1;
    }
    pfree(in);
    return 0;
}

static int
load_setup(starter_config_t *cfg, config_parsed_t *cfgp,
	char **perr)
{
    u_int err = 0;
    kw_list_t *kw;

    for (kw = cfgp->config_setup; kw; kw = kw->next)
    {
        char *p;
	kw_token_t token = kw->entry->token;
 
	if (token < KW_INTERFACES || token > KW_PKCS11PROXY)
	{
	    ERR_FOUND("unsupported keyword '%s' in config setup"
		     , kw->entry->name);
	    continue;
	}

	p = (char *)cfg + token_info[token].offset;

	switch (token_info[token].type)
	{
	case ARG_NONE:
	    ERR_FOUND("option '%s' not supported yet", kw->entry->name);
	    break;
	case ARG_BOOL:
	    {
		bool *b = (bool *)p;
		KW_BOOL(*b)
	    }
	    break;
	case ARG_UINT:
	    {
		u_int *u = (u_int *)p; 
		KW_INT(*u)
	    }
	    break;
	case ARG_DUR:
	    {
		u_int *u = (u_int *)p; 
		KW_DURATION(*u)
	    }
	    break;
	case ARG_STR:
	    {
		char **c = (char **)p;
		KW_STR(*c)
	    }
	    break;
	case ARG_STRC:
	    KW_STR_IN(cfg->setup.packetdefault, "drop reject pass")
	    break;
	case ARG_LST:
	    KW_LIST(cfg->setup.interfaces)
	    break;
	case ARG_LSTC:
	    if (token == KW_KLIPSDEBUG)
	    {
		KW_LIST_IN(cfg->setup.klipsdebug,
		    "tunnel tunnel-xmit pfkey xform eroute spi radij esp ah "
		    "ipcomp verbose all none")
	    }
	    else
	    {
		KW_LIST_IN(cfg->setup.plutodebug,
		    "none all raw crypt parsing emitting control lifecycle "
		    "klips dns natt oppo controlmore private")
	    }
	    break;
	default:
	    break;
	}
    }
    return err;
}

static int
kw_end(starter_conn_t *conn, starter_end_t *end, kw_token_t token
, char *value, char *name, char *conn_name
, defaultroute_t *defaultroute, char **perr)
{
    unsigned int err = 0;
    err_t er;
    int has_port_wildcard;        /* set if port is %any */

    switch (token)
    {
    case KW_HOST:
	if (streq(value, "%defaultroute"))
	{
	    if (defaultroute->defined)
	    {
		end->addr = defaultroute->addr;
		end->nexthop = defaultroute->nexthop;
	    }
	    else
		ERR_FOUND("default route not known: %s=%s", name, value);
	}
	else if (streq(value,"%any"))
        {
	    anyaddr(conn->addr_family, &end->addr);
	}
	else if (value[0] == '%')
	{
	    if (end->iface)
		pfree(end->iface);
	    end->iface = clone_str(value+1, "iface");
	    if (starter_iface_find(end->iface, conn->addr_family, &end->addr,
		&end->nexthop) == -1)
	    {
		conn->state = STATE_INVALID;
	    }
	}
	else
	{
	    er = ttoaddr(value, 0, conn->addr_family, &end->addr);
	    if (er)
		ERR_FOUND("bad addr: %s=%s [%s]", name, value, er);
	}
	break;
    case KW_NEXTHOP:
	if (streq(value, "%defaultroute"))
	{
	    if (defaultroute->defined)
		end->nexthop = defaultroute->nexthop;
	    else
		ERR_FOUND("default route not known: %s=%s", name, value);
	}
	else if (streq(value, "%direct"))
	    er = anyaddr(conn->addr_family, &end->nexthop);
	else
	    er = ttoaddr(value, 0, conn->addr_family, &end->nexthop);
	if (er)
	    ERR_FOUND("bad addr: %s=%s [%s]", name, value, er);
	break;
    case KW_SUBNET:
#ifdef VIRTUAL_IP
	if ((strlen(value) >= 6 && strncmp(value,"vhost:",6) == 0)
	||  (strlen(value) >= 5 && strncmp(value,"vnet:",5) == 0))
	{
	    er = NULL;
	    end->virt = clone_str(value, "virt");
	    if (!end->virt)
		ERR_FOUND("can't alloc memory");
	}
	else
	{
	    end->has_client = TRUE;
	    er = ttosubnet(value, 0, conn->tunnel_addr_family, &end->subnet);
	}
#else
	end->has_client = TRUE;
	end->has_client_wildcard = FALSE;
	er = ttosubnet(value, 0, conn->tunnel_addr_family, &end->subnet);
#endif
	if (er)
	    ERR_FOUND("bad subnet: %s=%s [%s]", name, value, er);
	break;
    case KW_SUBNETWITHIN:
	end->has_client = TRUE;
	end->has_client_wildcard = TRUE;
	er = ttosubnet(value, 0, conn->tunnel_addr_family, &end->subnet);
	break;
    case KW_PROTOPORT:
	er = ttoprotoport(value, 0, &end->protocol, &end->port, &has_port_wildcard);
	end->has_port_wildcard = has_port_wildcard;
	break;
    case KW_SOURCEIP:
	if (streq(value, "%modeconfig") || streq(value, "%modecfg"))
	{
	    end->modecfg = TRUE;
	}
	else
	{
	    er = ttoaddr(value, 0, conn->addr_family, &end->srcip);
	    if (er)
		ERR_FOUND("bad addr: %s=%s [%s]", name, value, er);
	    end->has_srcip = TRUE;
	}
	conn->policy |= POLICY_TUNNEL;
	break;
    case KW_UPDOWN:
	if (end->updown)
	    pfree(end->updown);
	end->updown = clone_str(value, "updown");
	break;
    case KW_ID:
	if (end->id)
	    pfree(end->id);
	end->id = clone_str(value, "id");
	break;
    case KW_RSASIGKEY:
	if (end->rsakey)
	   pfree(end->rsakey);
	end->rsakey = clone_str(value, "rsakey");
	break;
    case KW_CERT:
	if (end->cert)
	    pfree(end->cert);
	end->cert = clone_str(value, "cert");
	break;
    case KW_SENDCERT:
	if (streq(value, "yes") || streq(value, "always"))
	{
	    end->sendcert = CERT_ALWAYS_SEND;
	}
	else if (streq(value, "no") || streq(value, "never"))
	{
	    end->sendcert = CERT_NEVER_SEND;
	}
	else if (streq(value, "ifasked"))
	{
	    end->sendcert = CERT_SEND_IF_ASKED;
	}
	else
	{
	    plog("unknown sendcert value '%s' in conn %s", value, conn_name);
	}
	break;
    case KW_CA:
	if (end->ca)
	    pfree(end->ca);
	end->ca = clone_str(value, "ca");
	break;
    case KW_GROUPS:
	if (end->groups)
	    pfree(end->groups);
	end->groups = clone_str(value, "groups");
	break;
    default:
	ERR_FOUND("ignoring unknown keyword '%s' in conn %s", name, conn_name);
    }

    return err;
}

static int
load_conn(starter_conn_t *conn, config_parsed_t *cfgp
, kw_list_t *kw, defaultroute_t *defaultroute, char **perr)
{
    u_int err = 0;
	
    for ( ; kw; kw = kw->next)
    {
	kw_token_t token = kw->entry->token;

	if (token >= KW_LEFT && token <= KW_LEFTGROUPS)
	{
	    err += kw_end(conn, &conn->left, token - KW_LEFT + KW_HOST
			, kw->value, kw->entry->name, conn->name
			, defaultroute, perr);
	    continue;
	}
	else if (token >= KW_RIGHT && token <= KW_RIGHTGROUPS)
	{
	    err += kw_end(conn, &conn->right, token - KW_RIGHT + KW_HOST
			, kw->value, kw->entry->name, conn->name
			, defaultroute, perr);
	    continue;
	}

	switch (token)
	{
	case KW_KEYEXCHANGE:
	    if (!streq(kw->value, "ike"))
		ERR_FOUND("bad keyexchange value: %s=%s", kw->entry->name, kw->value);
	    break;
	case KW_AUTO:
	    if (streq(kw->value, "add"))
		conn->startup = STARTUP_ADD;
	    else if (streq(kw->value, "start"))
		conn->startup = STARTUP_START;
	    else if (streq(kw->value, "route"))
		conn->startup = STARTUP_ROUTE;
	    else if (streq(kw->value, "ignore"))
		conn->startup = STARTUP_NO;
	    else
		ERR_FOUND("bad auto value: %s=%s", kw->entry->name, kw->value);
	    break;
	case KW_TYPE:
	    KW_POLICY_FLAG("tunnel", "transport", POLICY_TUNNEL)
	    break;
	case KW_PFS:
	    KW_POLICY_FLAG("yes", "no", POLICY_PFS)
	    break;
	case KW_COMPRESS:
	    KW_POLICY_FLAG("yes", "no", POLICY_COMPRESS)
	    break; 
	case KW_AUTHBY:
	    KW_POLICY_FLAG_BI("rsasig", "secret", POLICY_RSASIG,POLICY_PSK)
	    break;
	case KW_KEYLIFE:
	    KW_DURATION(conn->sa_ipsec_life_seconds)
	    break;
	case KW_REKEYMARGIN:
	    KW_DURATION(conn->sa_rekey_margin)
	    break;
	case KW_IKELIFETIME:
	    KW_DURATION(conn->sa_ike_life_seconds)
	    break;
	case KW_KEYINGTRIES:
	    KW_INT(conn->sa_keying_tries)
	    break;
	case KW_REKEYFUZZ:
	    KW_PERCENT(conn->sa_rekey_fuzz)
	    break;
	case KW_REKEY:
	    KW_POLICY_FLAG("no", "yes", POLICY_DONT_REKEY)
	    break;
	case KW_ESP:
	    KW_STR(conn->esp)
	    break;
	case KW_IKE:
	    KW_STR(conn->ike)
	    break;
	case KW_PFSGROUP:
	    KW_STR_IN(conn->pfsgroup
		, "modp1024 modp1536 modp2048 modp3072 modp4096 modp6144 modp8192")
	    break;
	case KW_DPDDELAY:
	    KW_DURATION(conn->dpd_delay)
	    break;
	case KW_DPDTIMEOUT:
	    KW_DURATION(conn->dpd_timeout)
	    break;
	case KW_DPDACTION:
	    if (streq(kw->value, "none"))
		conn->dpd_action = DPD_ACTION_NONE;
	    else if (streq(kw->value, "clear"))
		conn->dpd_action = DPD_ACTION_CLEAR;
	    else if (streq(kw->value, "hold"))
		conn->dpd_action = DPD_ACTION_HOLD;
	    else if (streq(kw->value, "restart"))
		conn->dpd_action = DPD_ACTION_RESTART;
	    break;
	default:
	    ERR_FOUND("unsupported keyword '%s' in conn %s"
		     , kw->entry->name, conn->name);
	}
    }
    return err;
}

static void
conn_default(char *name, starter_conn_t *conn, starter_conn_t *def)
{
    starter_end_t *end;

    memcpy(conn, def, sizeof(starter_conn_t));
    conn->name = clone_str(name, "conn name");

    for (end = &conn->left; ; end = &conn->right)
    {
	end->iface  = clone_str(end->iface,  "default iface");
	end->id     = clone_str(end->id,     "default id");
	end->rsakey = clone_str(end->rsakey, "default rsakey");
	end->updown = clone_str(end->updown, "default update");
	end->cert   = clone_str(end->cert,   "default cert");
	end->ca     = clone_str(end->ca,     "default ca");
	end->groups = clone_str(end->groups, "default groups");
#ifdef VIRTUAL_IP
	end->virt   = clone_str(end->virt,   "default virt");
#endif
	if (end == &conn->right)
	    break;
    }
    conn->esp      = clone_str(conn->esp,      "default esp");
    conn->ike      = clone_str(conn->ike,      "default ike");
    conn->pfsgroup = clone_str(conn->pfsgroup, "default pfsgroup");
}

static int
load_ca(starter_ca_t *ca, config_parsed_t *cfgp,
	kw_list_t *kw, char **perr)
{
    u_int err = 0;
	
    for ( ; kw; kw = kw->next)
    {
	switch (kw->entry->token)
	{
	case KW_AUTO:
	    if (streq(kw->value, "add"))
		ca->startup = STARTUP_ADD;
	    else if (streq(kw->value, "start"))
		ca->startup = STARTUP_ADD;
	    else if (streq(kw->value, "ignore"))
		ca->startup = STARTUP_NO;
	    else
		ERR_FOUND("bad auto value: %s=%s", kw->entry->name, kw->value);
	    break;
	case KW_STRICTCRLPOLICY:
	    KW_BOOL(ca->strict)
	    break;
	case KW_CACERT:
	    KW_STR(ca->cacert)
	    break;
	case KW_LDAPHOST:
	    KW_STR(ca->ldaphost)
	    break;
	case KW_LDAPBASE:
	    KW_STR(ca->ldapbase)
	    break;
	case KW_CRLURI:
	    KW_STR(ca->crluri)
	    break;
	case KW_CRLURI2:
	    KW_STR(ca->crluri2)
	    break;
	case KW_OCSPURI:
	    KW_STR(ca->ocspuri)
	    break;
	default:
	    ERR_FOUND("unsupported keyword '%s' in ca %s"
		     , kw->entry->name, ca->name);
	}
    }

    return err;
}

static void
ca_default(char *name, starter_ca_t *ca, starter_ca_t *def)
{
    memcpy(ca, def, sizeof(starter_ca_t));
    ca->name = clone_str(name, "ca name");

    ca->cacert   = clone_str(ca->cacert,   "default cacert");
    ca->ldaphost = clone_str(ca->ldaphost, "default ldaphost");
    ca->ldapbase = clone_str(ca->ldapbase, "default ldapbase");
    ca->crluri   = clone_str(ca->crluri,   "default crluri");
    ca->crluri2  = clone_str(ca->crluri2,  "default crluri2");
    ca->ocspuri  = clone_str(ca->ocspuri,  "default ocspuri");
}

#undef DUP_STR

starter_config_t *
confread_load(const char *file, char **perr)
{
    starter_config_t *cfg = NULL;
    config_parsed_t  *cfgp;
    section_list_t   *sconn, *sca;
    starter_conn_t   *conn;
    starter_ca_t     *ca;

    u_int err = 0;

    /* load IPSec configuration file  */
    cfgp = parser_load_conf(file, perr);
    if (!cfgp)
	return NULL;

    cfg = (starter_config_t *)alloc_thing(starter_config_t, "starter_config_t");
    if (!cfg)
    {
	if (perr)
	    *perr = strdup("can't allocate mem in confread_load()");
	parser_free_conf(cfgp);
	return NULL;
    }

    /* set default values */
    default_values(cfg);

    /* determine default route */
    get_defaultroute(&cfg->defaultroute);
 
    /* load config setup section */
    err += load_setup(cfg, cfgp, perr);

     /* find %default ca section */
    for (sca = cfgp->ca_first; !err && sca; sca = sca->next)
    {
	if (streq(sca->name, "%default"))
	{
	    DBG(DBG_CONTROL,
		DBG_log("Loading default ca")
	    )
	    err += load_ca(&cfg->ca_default, cfgp, sca->kw, perr);
	}
    }

    /* load other ca sections */
    for (sca = cfgp->ca_first; !err && sca; sca = sca->next)
    {
	if (streq(sca->name, "%default"))
	    continue;
	DBG(DBG_CONTROL,
	    DBG_log("Loading ca %s", sca->name)
	)
	ca = (starter_ca_t *)alloc_thing(starter_ca_t, "starter_ca_t");
	if (!ca)
	{
	    if (perr)
		*perr = strdup("can't allocate mem in confread_load()");
	    parser_free_conf(cfgp);
	    confread_free(cfg);
	    return NULL;
	}
	ca_default(sca->name, ca, &cfg->ca_default);
	ca->next = NULL;
	if (cfg->ca_last)
	    cfg->ca_last->next = ca;
	cfg->ca_last = ca;
	if (!cfg->ca_first)
	    cfg->ca_first = ca;
	err += load_ca(ca, cfgp, sca->kw, perr);
	if (ca->startup != STARTUP_NO)
	    ca->state = STATE_TO_ADD;
    }

    /* find %default conn section */
    for (sconn = cfgp->conn_first; !err && sconn; sconn = sconn->next)
    {
	if (streq(sconn->name, "%default"))
	{
	    DBG(DBG_CONTROL,
		DBG_log("Loading default conn")
	    )
	    err += load_conn(&cfg->conn_default, cfgp, sconn->kw
			    , &cfg->defaultroute, perr);
	}
    }

    /* load other conn sections */
    for (sconn = cfgp->conn_first; !err && sconn; sconn = sconn->next)
    {
	if (streq(sconn->name, "%default"))
	    continue;
	DBG(DBG_CONTROL,
	    DBG_log("Loading conn %s", sconn->name)
	)
	conn = (starter_conn_t *)alloc_thing(starter_conn_t, "starter_conn_t");
	if (!conn)
	{
	    if (perr)
		*perr = strdup("can't allocate mem in confread_load()");
	    parser_free_conf(cfgp);
	    confread_free(cfg);
	    return NULL;
	}
	conn_default(sconn->name, conn, &cfg->conn_default);
	conn->next = NULL;
	if (cfg->conn_last)
	    cfg->conn_last->next = conn;
	cfg->conn_last = conn;
	if (!cfg->conn_first)
	    cfg->conn_first = conn;
	err += load_conn(conn, cfgp, sconn->kw, &cfg->defaultroute, perr);
	if (conn->startup != STARTUP_NO)
	    conn->state = STATE_TO_ADD;
    }

    parser_free_conf(cfgp);

    if (err)
    {
	confread_free(cfg);
	cfg = NULL;
    }

    return cfg;
}

static void
confread_free_conn(starter_conn_t *conn)
{
    starter_end_t *end;

    for (end = &conn->left; ; end = &conn->right)
    {
	FREE_STR(end->iface);
	FREE_STR(end->id);
	FREE_STR(end->rsakey);
	FREE_STR(end->updown);
	FREE_STR(end->cert);
	FREE_STR(end->ca);
	FREE_STR(end->groups);
#ifdef VIRTUAL_IP
	FREE_STR(end->virt);
#endif
	if (end == &conn->right)
	    break;
    }
    FREE_STR(conn->esp);
    FREE_STR(conn->ike);
    FREE_STR(conn->pfsgroup);
    FREE_STR(conn->name);
}

static void
confread_free_ca(starter_ca_t *ca)
{
    FREE_STR(ca->cacert);
    FREE_STR(ca->ldaphost);
    FREE_STR(ca->ldapbase);
    FREE_STR(ca->crluri);
    FREE_STR(ca->crluri2);
    FREE_STR(ca->ocspuri);
    FREE_STR(ca->name);
}

void 
confread_free(starter_config_t *cfg)
{
    starter_conn_t *conn = cfg->conn_first;
    starter_ca_t   *ca   = cfg->ca_first;

    FREE_LST(cfg->setup.interfaces);
    FREE_LST(cfg->setup.klipsdebug);
    FREE_LST(cfg->setup.plutodebug);
    FREE_STR(cfg->setup.prepluto);
    FREE_STR(cfg->setup.postpluto);
    FREE_STR(cfg->setup.packetdefault);
    FREE_STR(cfg->setup.pkcs11module);
#ifdef VIRTUAL_IP
    FREE_STR(cfg->setup.virtual_private);
#endif

    confread_free_conn(&cfg->conn_default);

    while (conn != NULL)
    {
	starter_conn_t *conn_aux = conn;

	conn = conn->next;
	confread_free_conn(conn_aux);
	pfree(conn_aux);
    }

    confread_free_ca(&cfg->ca_default);

    while (ca != NULL)
    {
	starter_ca_t *ca_aux = ca;

	ca = ca->next;
	confread_free_ca(ca_aux);
	pfree(ca_aux);
    }

    pfree(cfg);
}

#undef FREE_STR

